@@ -10,11 +10,11 @@ import json
 import mimetypes
 import os
 import re
-import traceback
 import types
 import warnings
 from http.client import responses
-from typing import TYPE_CHECKING, Awaitable
+from logging import Logger
+from typing import TYPE_CHECKING, Any, Awaitable, Sequence, cast
 from urllib.parse import urlparse
 
 import prometheus_client
@@ -26,9 +26,11 @@ from tornado.log import app_log
 from traitlets.config import Application
 
 import jupyter_server
+from jupyter_server import CallContext
 from jupyter_server._sysinfo import get_sys_info
 from jupyter_server._tz import utcnow
-from jupyter_server.auth import authorized
+from jupyter_server.auth.decorator import authorized
+from jupyter_server.auth.identity import User
 from jupyter_server.i18n import combine_translations
 from jupyter_server.services.security import csp_report_uri
 from jupyter_server.utils import (
@@ -41,7 +43,17 @@ from jupyter_server.utils import (
 )
 
 if TYPE_CHECKING:
-    from jupyter_server.auth.identity import User
+    from jupyter_client.kernelspec import KernelSpecManager
+    from jupyter_server_terminals.terminalmanager import TerminalManager
+    from tornado.concurrent import Future
+
+    from jupyter_server.auth.authorizer import Authorizer
+    from jupyter_server.auth.identity import IdentityProvider
+    from jupyter_server.serverapp import ServerApp
+    from jupyter_server.services.config.manager import ConfigManager
+    from jupyter_server.services.contents.manager import ContentsManager
+    from jupyter_server.services.kernels.kernelmanager import AsyncMappingKernelManager
+    from jupyter_server.services.sessions.sessionmanager import SessionManager
 
 # -----------------------------------------------------------------------------
 # Top-level handlers
@@ -52,16 +64,16 @@ _sys_info_cache = None
 
 def json_sys_info():
     """Get sys info as json."""
-    global _sys_info_cache
+    global _sys_info_cache  # noqa: PLW0603
     if _sys_info_cache is None:
         _sys_info_cache = json.dumps(get_sys_info())
     return _sys_info_cache
 
 
-def log():
+def log() -> Logger:
     """Get the application log."""
     if Application.initialized():
-        return Application.instance().log
+        return cast(Logger, Application.instance().log)
     else:
         return app_log
 
@@ -71,17 +83,17 @@ class AuthenticatedHandler(web.RequestHandler):
 
     @property
     def base_url(self) -> str:
-        return self.settings.get("base_url", "/")
+        return cast(str, self.settings.get("base_url", "/"))
 
     @property
-    def content_security_policy(self):
+    def content_security_policy(self) -> str:
         """The default Content-Security-Policy header
 
         Can be overridden by defining Content-Security-Policy in settings['headers']
         """
         if "Content-Security-Policy" in self.settings.get("headers", {}):
             # user-specified, don't override
-            return self.settings["headers"]["Content-Security-Policy"]
+            return cast(str, self.settings["headers"]["Content-Security-Policy"])
 
         return "; ".join(
             [
@@ -92,7 +104,7 @@ class AuthenticatedHandler(web.RequestHandler):
             ]
         )
 
-    def set_default_headers(self):
+    def set_default_headers(self) -> None:
         """Set the default headers."""
         headers = {}
         headers["X-Content-Type-Options"] = "nosniff"
@@ -113,7 +125,7 @@ class AuthenticatedHandler(web.RequestHandler):
                 )
 
     @property
-    def cookie_name(self):
+    def cookie_name(self) -> str:
         warnings.warn(
             """JupyterHandler.login_handler is deprecated in 2.0,
             use JupyterHandler.identity_provider.
@@ -123,7 +135,7 @@ class AuthenticatedHandler(web.RequestHandler):
         )
         return self.identity_provider.get_cookie_name(self)
 
-    def force_clear_cookie(self, name, path="/", domain=None):
+    def force_clear_cookie(self, name: str, path: str = "/", domain: str | None = None) -> None:
         """Force a cookie clear."""
         warnings.warn(
             """JupyterHandler.login_handler is deprecated in 2.0,
@@ -132,9 +144,9 @@ class AuthenticatedHandler(web.RequestHandler):
             DeprecationWarning,
             stacklevel=2,
         )
-        return self.identity_provider._force_clear_cookie(self, name, path=path, domain=domain)
+        self.identity_provider._force_clear_cookie(self, name, path=path, domain=domain)
 
-    def clear_login_cookie(self):
+    def clear_login_cookie(self) -> None:
         """Clear a login cookie."""
         warnings.warn(
             """JupyterHandler.login_handler is deprecated in 2.0,
@@ -143,9 +155,9 @@ class AuthenticatedHandler(web.RequestHandler):
             DeprecationWarning,
             stacklevel=2,
         )
-        return self.identity_provider.clear_login_cookie(self)
+        self.identity_provider.clear_login_cookie(self)
 
-    def get_current_user(self):
+    def get_current_user(self) -> str:
         """Get the current user."""
         clsname = self.__class__.__name__
         msg = (
@@ -159,11 +171,11 @@ class AuthenticatedHandler(web.RequestHandler):
                 DeprecationWarning,
                 stacklevel=2,
             )
-            return self._jupyter_current_user
+            return cast(str, self._jupyter_current_user)
         # haven't called get_user in prepare, raise
         raise RuntimeError(msg)
 
-    def skip_check_origin(self):
+    def skip_check_origin(self) -> bool:
         """Ask my login_handler if I should skip the origin_check
 
         For example: in the default LoginHandler, if a request is token-authenticated,
@@ -175,18 +187,18 @@ class AuthenticatedHandler(web.RequestHandler):
         return not self.identity_provider.should_check_origin(self)
 
     @property
-    def token_authenticated(self):
+    def token_authenticated(self) -> bool:
         """Have I been authenticated with a token?"""
         return self.identity_provider.is_token_authenticated(self)
 
     @property
-    def logged_in(self):
+    def logged_in(self) -> bool:
         """Is a user currently logged in?"""
         user = self.current_user
-        return user and user != "anonymous"
+        return bool(user and user != "anonymous")
 
     @property
-    def login_handler(self):
+    def login_handler(self) -> Any:
         """Return the login handler for this application, if any."""
         warnings.warn(
             """JupyterHandler.login_handler is deprecated in 2.0,
@@ -198,22 +210,22 @@ class AuthenticatedHandler(web.RequestHandler):
         return self.identity_provider.login_handler_class
 
     @property
-    def token(self):
+    def token(self) -> str | None:
         """Return the login token for this application, if any."""
         return self.identity_provider.token
 
     @property
-    def login_available(self):
+    def login_available(self) -> bool:
         """May a user proceed to log in?
 
         This returns True if login capability is available, irrespective of
         whether the user is already logged in or not.
 
         """
-        return self.identity_provider.login_available
+        return cast(bool, self.identity_provider.login_available)
 
     @property
-    def authorizer(self):
+    def authorizer(self) -> Authorizer:
         if "authorizer" not in self.settings:
             warnings.warn(
                 "The Tornado web application does not have an 'authorizer' defined "
@@ -224,6 +236,7 @@ class AuthenticatedHandler(web.RequestHandler):
                 "https://github.com/jupyter-server/jupyter_server/blob/"
                 "653740cbad7ce0c8a8752ce83e4d3c2c754b13cb/jupyter_server/serverapp.py"
                 "#L234-L256",
+                stacklevel=2,
             )
             from jupyter_server.auth import AllowAllAuthorizer
 
@@ -232,10 +245,10 @@ class AuthenticatedHandler(web.RequestHandler):
                 identity_provider=self.identity_provider,
             )
 
-        return self.settings.get("authorizer")
+        return cast("Authorizer", self.settings.get("authorizer"))
 
     @property
-    def identity_provider(self):
+    def identity_provider(self) -> IdentityProvider:
         if "identity_provider" not in self.settings:
             warnings.warn(
                 "The Tornado web application does not have an 'identity_provider' defined "
@@ -245,6 +258,7 @@ class AuthenticatedHandler(web.RequestHandler):
                 "add an identity provider to the tornado settings: "
                 "https://github.com/jupyter-server/jupyter_server/blob/v2.0.0/"
                 "jupyter_server/serverapp.py#L242",
+                stacklevel=2,
             )
             from jupyter_server.auth import IdentityProvider
 
@@ -252,7 +266,7 @@ class AuthenticatedHandler(web.RequestHandler):
             self.settings["identity_provider"] = IdentityProvider(
                 config=self.settings.get("config", None)
             )
-        return self.settings["identity_provider"]
+        return cast("IdentityProvider", self.settings["identity_provider"])
 
 
 class JupyterHandler(AuthenticatedHandler):
@@ -262,115 +276,115 @@ class JupyterHandler(AuthenticatedHandler):
     """
 
     @property
-    def config(self):
-        return self.settings.get("config", None)
+    def config(self) -> dict[str, Any] | None:
+        return cast("dict[str, Any] | None", self.settings.get("config", None))
 
     @property
-    def log(self):
+    def log(self) -> Logger:
         """use the Jupyter log by default, falling back on tornado's logger"""
         return log()
 
     @property
-    def jinja_template_vars(self):
+    def jinja_template_vars(self) -> dict[str, Any]:
         """User-supplied values to supply to jinja templates."""
-        return self.settings.get("jinja_template_vars", {})
+        return cast("dict[str, Any]", self.settings.get("jinja_template_vars", {}))
 
     @property
-    def serverapp(self):
-        return self.settings["serverapp"]
+    def serverapp(self) -> ServerApp | None:
+        return cast("ServerApp | None", self.settings["serverapp"])
 
     # ---------------------------------------------------------------
     # URLs
     # ---------------------------------------------------------------
 
     @property
-    def version_hash(self):
+    def version_hash(self) -> str:
         """The version hash to use for cache hints for static files"""
-        return self.settings.get("version_hash", "")
+        return cast(str, self.settings.get("version_hash", ""))
 
     @property
-    def mathjax_url(self):
-        url = self.settings.get("mathjax_url", "")
+    def mathjax_url(self) -> str:
+        url = cast(str, self.settings.get("mathjax_url", ""))
         if not url or url_is_absolute(url):
             return url
         return url_path_join(self.base_url, url)
 
     @property
-    def mathjax_config(self):
-        return self.settings.get("mathjax_config", "TeX-AMS-MML_HTMLorMML-full,Safe")
+    def mathjax_config(self) -> str:
+        return cast(str, self.settings.get("mathjax_config", "TeX-AMS-MML_HTMLorMML-full,Safe"))
 
     @property
-    def default_url(self):
-        return self.settings.get("default_url", "")
+    def default_url(self) -> str:
+        return cast(str, self.settings.get("default_url", ""))
 
     @property
-    def ws_url(self):
-        return self.settings.get("websocket_url", "")
+    def ws_url(self) -> str:
+        return cast(str, self.settings.get("websocket_url", ""))
 
     @property
-    def contents_js_source(self):
+    def contents_js_source(self) -> str:
         self.log.debug(
             "Using contents: %s",
             self.settings.get("contents_js_source", "services/contents"),
         )
-        return self.settings.get("contents_js_source", "services/contents")
+        return cast(str, self.settings.get("contents_js_source", "services/contents"))
 
     # ---------------------------------------------------------------
     # Manager objects
     # ---------------------------------------------------------------
 
     @property
-    def kernel_manager(self):
-        return self.settings["kernel_manager"]
+    def kernel_manager(self) -> AsyncMappingKernelManager:
+        return cast("AsyncMappingKernelManager", self.settings["kernel_manager"])
 
     @property
-    def contents_manager(self):
-        return self.settings["contents_manager"]
+    def contents_manager(self) -> ContentsManager:
+        return cast("ContentsManager", self.settings["contents_manager"])
 
     @property
-    def session_manager(self):
-        return self.settings["session_manager"]
+    def session_manager(self) -> SessionManager:
+        return cast("SessionManager", self.settings["session_manager"])
 
     @property
-    def terminal_manager(self):
-        return self.settings["terminal_manager"]
+    def terminal_manager(self) -> TerminalManager:
+        return cast("TerminalManager", self.settings["terminal_manager"])
 
     @property
-    def kernel_spec_manager(self):
-        return self.settings["kernel_spec_manager"]
+    def kernel_spec_manager(self) -> KernelSpecManager:
+        return cast("KernelSpecManager", self.settings["kernel_spec_manager"])
 
     @property
-    def config_manager(self):
-        return self.settings["config_manager"]
+    def config_manager(self) -> ConfigManager:
+        return cast("ConfigManager", self.settings["config_manager"])
 
     @property
     def event_logger(self) -> EventLogger:
-        return self.settings["event_logger"]
+        return cast("EventLogger", self.settings["event_logger"])
 
     # ---------------------------------------------------------------
     # CORS
     # ---------------------------------------------------------------
 
     @property
-    def allow_origin(self):
+    def allow_origin(self) -> str:
         """Normal Access-Control-Allow-Origin"""
-        return self.settings.get("allow_origin", "")
+        return cast(str, self.settings.get("allow_origin", ""))
 
     @property
-    def allow_origin_pat(self):
+    def allow_origin_pat(self) -> str | None:
         """Regular expression version of allow_origin"""
-        return self.settings.get("allow_origin_pat", None)
+        return cast("str | None", self.settings.get("allow_origin_pat", None))
 
     @property
-    def allow_credentials(self):
+    def allow_credentials(self) -> bool:
         """Whether to set Access-Control-Allow-Credentials"""
-        return self.settings.get("allow_credentials", False)
+        return cast(bool, self.settings.get("allow_credentials", False))
 
-    def set_default_headers(self):
+    def set_default_headers(self) -> None:
         """Add CORS headers, if defined"""
         super().set_default_headers()
 
-    def set_cors_headers(self):
+    def set_cors_headers(self) -> None:
         """Add CORS headers, if defined
 
         Now that current_user is async (jupyter-server 2.0),
@@ -392,7 +406,7 @@ class JupyterHandler(AuthenticatedHandler):
         if self.allow_credentials:
             self.set_header("Access-Control-Allow-Credentials", "true")
 
-    def set_attachment_header(self, filename):
+    def set_attachment_header(self, filename: str) -> None:
         """Set Content-Disposition: attachment header
 
         As a method to ensure handling of filename encoding
@@ -400,13 +414,10 @@ class JupyterHandler(AuthenticatedHandler):
         escaped_filename = url_escape(filename)
         self.set_header(
             "Content-Disposition",
-            "attachment;"
-            " filename*=utf-8''{utf8}".format(
-                utf8=escaped_filename,
-            ),
+            f"attachment; filename*=utf-8''{escaped_filename}",
         )
 
-    def get_origin(self):
+    def get_origin(self) -> str | None:
         # Handle WebSocket Origin naming convention differences
         # The difference between version 8 and 13 is that in 8 the
         # client sends a "Sec-Websocket-Origin" header and in 13 it's
@@ -419,7 +430,7 @@ class JupyterHandler(AuthenticatedHandler):
 
     # origin_to_satisfy_tornado is present because tornado requires
     # check_origin to take an origin argument, but we don't use it
-    def check_origin(self, origin_to_satisfy_tornado=""):
+    def check_origin(self, origin_to_satisfy_tornado: str = "") -> bool:
         """Check Origin for cross-site API requests, including websockets
 
         Copied from WebSocket with changes:
@@ -451,7 +462,7 @@ class JupyterHandler(AuthenticatedHandler):
 
         # Check CORS headers
         if self.allow_origin:
-            allow = self.allow_origin == origin
+            allow = bool(self.allow_origin == origin)
         elif self.allow_origin_pat:
             allow = bool(re.match(self.allow_origin_pat, origin))
         else:
@@ -466,7 +477,7 @@ class JupyterHandler(AuthenticatedHandler):
             )
         return allow
 
-    def check_referer(self):
+    def check_referer(self) -> bool:
         """Check Referer for cross-site requests.
         Disables requests to certain endpoints with
         external or missing Referer.
@@ -512,15 +523,15 @@ class JupyterHandler(AuthenticatedHandler):
             )
         return allow
 
-    def check_xsrf_cookie(self):
+    def check_xsrf_cookie(self) -> None:
         """Bypass xsrf cookie checks when token-authenticated"""
         if not hasattr(self, "_jupyter_current_user"):
             # Called too early, will be checked later
-            return
+            return None
         if self.token_authenticated or self.settings.get("disable_check_xsrf", False):
             # Token-authenticated requests do not need additional XSRF-check
             # Servers without authentication are vulnerable to XSRF
-            return
+            return None
         try:
             return super().check_xsrf_cookie()
         except web.HTTPError as e:
@@ -536,7 +547,7 @@ class JupyterHandler(AuthenticatedHandler):
             else:
                 raise
 
-    def check_host(self):
+    def check_host(self) -> bool:
         """Check the host header if remote access disallowed.
 
         Returns True if the request should continue, False otherwise.
@@ -578,8 +589,11 @@ class JupyterHandler(AuthenticatedHandler):
             )
         return allow
 
-    async def prepare(self):
-        """Pepare a response."""
+    async def prepare(self) -> Awaitable[None] | None:  # type:ignore[override]
+        """Prepare a response."""
+        # Set the current Jupyter Handler context variable.
+        CallContext.set(CallContext.JUPYTER_HANDLER, self)
+
         if not self.check_host():
             self.current_user = self._jupyter_current_user = None
             raise web.HTTPError(403)
@@ -598,9 +612,9 @@ class JupyterHandler(AuthenticatedHandler):
                 "Overriding JupyterHandler.get_current_user is deprecated in jupyter-server 2.0."
                 " Use an IdentityProvider class.",
                 DeprecationWarning,
-                # stacklevel not useful here
+                stacklevel=1,
             )
-            user = self.get_current_user()
+            user = User(self.get_current_user())
         else:
             _user = self.identity_provider.get_user(self)
             if isinstance(_user, Awaitable):
@@ -633,7 +647,7 @@ class JupyterHandler(AuthenticatedHandler):
         return template.render(**ns)
 
     @property
-    def template_namespace(self):
+    def template_namespace(self) -> dict[str, Any]:
         return dict(
             base_url=self.base_url,
             default_url=self.default_url,
@@ -656,7 +670,7 @@ class JupyterHandler(AuthenticatedHandler):
             **self.jinja_template_vars,
         )
 
-    def get_json_body(self):
+    def get_json_body(self) -> dict[str, Any] | None:
         """Return the body of the request as JSON data."""
         if not self.request.body:
             return None
@@ -668,9 +682,9 @@ class JupyterHandler(AuthenticatedHandler):
             self.log.debug("Bad JSON: %r", body)
             self.log.error("Couldn't parse JSON", exc_info=True)
             raise web.HTTPError(400, "Invalid JSON in body of request") from e
-        return model
+        return cast("dict[str, Any]", model)
 
-    def write_error(self, status_code, **kwargs):
+    def write_error(self, status_code: int, **kwargs: Any) -> None:
         """render custom error pages"""
         exc_info = kwargs.get("exc_info")
         message = ""
@@ -712,17 +726,17 @@ class JupyterHandler(AuthenticatedHandler):
 class APIHandler(JupyterHandler):
     """Base class for API handlers"""
 
-    async def prepare(self):
+    async def prepare(self) -> None:
         """Prepare an API response."""
         await super().prepare()
         if not self.check_origin():
             raise web.HTTPError(404)
 
-    def write_error(self, status_code, **kwargs):
+    def write_error(self, status_code: int, **kwargs: Any) -> None:
         """APIHandler errors are JSON, not human pages"""
         self.set_header("Content-Type", "application/json")
         message = responses.get(status_code, "Unknown HTTP Error")
-        reply: dict = {
+        reply: dict[str, Any] = {
             "message": message,
         }
         exc_info = kwargs.get("exc_info")
@@ -734,11 +748,13 @@ class APIHandler(JupyterHandler):
             else:
                 reply["message"] = "Unhandled error"
                 reply["reason"] = None
-                reply["traceback"] = "".join(traceback.format_exception(*exc_info))
+                # backward-compatibility: traceback field is present,
+                # but always empty
+                reply["traceback"] = ""
         self.log.warning("wrote error: %r", reply["message"], exc_info=True)
         self.finish(json.dumps(reply))
 
-    def get_login_url(self):
+    def get_login_url(self) -> str:
         """Get the login url."""
         # if get_login_url is invoked in an API handler,
         # that means @web.authenticated is trying to trigger a redirect.
@@ -748,7 +764,7 @@ class APIHandler(JupyterHandler):
         return super().get_login_url()
 
     @property
-    def content_security_policy(self):
+    def content_security_policy(self) -> str:
         csp = "; ".join(
             [
                 super().content_security_policy,
@@ -760,23 +776,25 @@ class APIHandler(JupyterHandler):
     # set _track_activity = False on API handlers that shouldn't track activity
     _track_activity = True
 
-    def update_api_activity(self):
+    def update_api_activity(self) -> None:
         """Update last_activity of API requests"""
         # record activity of authenticated requests
         if (
             self._track_activity
-            and getattr(self, "_user_cache", None)
+            and getattr(self, "_jupyter_current_user", None)
             and self.get_argument("no_track_activity", None) is None
         ):
             self.settings["api_last_activity"] = utcnow()
 
-    def finish(self, *args, **kwargs):
+    def finish(self, *args: Any, **kwargs: Any) -> Future[Any]:
         """Finish an API response."""
         self.update_api_activity()
-        self.set_header("Content-Type", "application/json")
+        # Allow caller to indicate content-type...
+        set_content_type = kwargs.pop("set_content_type", "application/json")
+        self.set_header("Content-Type", set_content_type)
         return super().finish(*args, **kwargs)
 
-    def options(self, *args, **kwargs):
+    def options(self, *args: Any, **kwargs: Any) -> None:
         """Get the options."""
         if "Access-Control-Allow-Headers" in self.settings.get("headers", {}):
             self.set_header(
@@ -819,7 +837,7 @@ class APIHandler(JupyterHandler):
 class Template404(JupyterHandler):
     """Render our 404 template"""
 
-    async def prepare(self):
+    async def prepare(self) -> None:
         """Prepare a 404 response."""
         await super().prepare()
         raise web.HTTPError(404)
@@ -831,29 +849,32 @@ class AuthenticatedFileHandler(JupyterHandler, web.StaticFileHandler):
     auth_resource = "contents"
 
     @property
-    def content_security_policy(self):
+    def content_security_policy(self) -> str:
         # In case we're serving HTML/SVG, confine any Javascript to a unique
         # origin so it can't interact with the Jupyter server.
         return super().content_security_policy + "; sandbox allow-scripts"
 
     @web.authenticated
     @authorized
-    def head(self, path):
+    def head(self, path: str) -> Awaitable[None]:  # type:ignore[override]
         """Get the head response for a path."""
         self.check_xsrf_cookie()
         return super().head(path)
 
     @web.authenticated
     @authorized
-    def get(self, path, **kwargs):
+    def get(  # type:ignore[override]
+        self, path: str, **kwargs: Any
+    ) -> Awaitable[None]:
         """Get a file by path."""
+        self.check_xsrf_cookie()
         if os.path.splitext(path)[1] == ".ipynb" or self.get_argument("download", None):
             name = path.rsplit("/", 1)[-1]
             self.set_attachment_header(name)
 
         return web.StaticFileHandler.get(self, path, **kwargs)
 
-    def get_content_type(self):
+    def get_content_type(self) -> str:
         """Get the content type."""
         assert self.absolute_path is not None
         path = self.absolute_path.strip("/")
@@ -870,18 +891,18 @@ class AuthenticatedFileHandler(JupyterHandler, web.StaticFileHandler):
             else:
                 return super().get_content_type()
 
-    def set_headers(self):
+    def set_headers(self) -> None:
         """Set the headers."""
         super().set_headers()
         # disable browser caching, rely on 304 replies for savings
         if "v" not in self.request.arguments:
             self.add_header("Cache-Control", "no-cache")
 
-    def compute_etag(self):
+    def compute_etag(self) -> str | None:
         """Compute the etag."""
         return None
 
-    def validate_absolute_path(self, root, absolute_path):
+    def validate_absolute_path(self, root: str, absolute_path: str) -> str:
         """Validate and return the absolute path.
 
         Requires tornado 3.1
@@ -891,7 +912,7 @@ class AuthenticatedFileHandler(JupyterHandler, web.StaticFileHandler):
         abs_path = super().validate_absolute_path(root, absolute_path)
         abs_root = os.path.abspath(root)
         assert abs_path is not None
-        if is_hidden(abs_path, abs_root) and not self.contents_manager.allow_hidden:
+        if not self.contents_manager.allow_hidden and is_hidden(abs_path, abs_root):
             self.log.info(
                 "Refusing to serve hidden file, via 404 Error, use flag 'ContentsManager.allow_hidden' to enable"
             )
@@ -899,7 +920,7 @@ class AuthenticatedFileHandler(JupyterHandler, web.StaticFileHandler):
         return abs_path
 
 
-def json_errors(method):  # pragma: no cover
+def json_errors(method: Any) -> Any:  # pragma: no cover
     """Decorate methods with this to return GitHub style JSON errors.
 
     This should be used on any JSON API on any handler method that can raise HTTPErrors.
@@ -934,37 +955,55 @@ HTTPError = web.HTTPError
 
 
 class FileFindHandler(JupyterHandler, web.StaticFileHandler):
-    """subclass of StaticFileHandler for serving files from a search path"""
+    """subclass of StaticFileHandler for serving files from a search path
+
+    The setting "static_immutable_cache" can be set up to serve some static
+    file as immutable (e.g. file name containing a hash). The setting is a
+    list of base URL, every static file URL starting with one of those will
+    be immutable.
+    """
 
     # cache search results, don't search for files more than once
-    _static_paths: dict = {}
-    root: tuple  # type:ignore[assignment]
+    _static_paths: dict[str, str] = {}
+    root: tuple[str]  # type:ignore[assignment]
 
-    def set_headers(self):
+    def set_headers(self) -> None:
         """Set the headers."""
         super().set_headers()
+
+        immutable_paths = self.settings.get("static_immutable_cache", [])
+
+        # allow immutable cache for files
+        if any(self.request.path.startswith(path) for path in immutable_paths):
+            self.set_header("Cache-Control", "public, max-age=31536000, immutable")
+
         # disable browser caching, rely on 304 replies for savings
-        if "v" not in self.request.arguments or any(
+        elif "v" not in self.request.arguments or any(
             self.request.path.startswith(path) for path in self.no_cache_paths
         ):
             self.set_header("Cache-Control", "no-cache")
 
-    def initialize(self, path, default_filename=None, no_cache_paths=None):
+    def initialize(
+        self,
+        path: str | list[str],
+        default_filename: str | None = None,
+        no_cache_paths: list[str] | None = None,
+    ) -> None:
         """Initialize the file find handler."""
         self.no_cache_paths = no_cache_paths or []
 
         if isinstance(path, str):
             path = [path]
 
-        self.root = tuple(os.path.abspath(os.path.expanduser(p)) + os.sep for p in path)
+        self.root = tuple(os.path.abspath(os.path.expanduser(p)) + os.sep for p in path)  # type:ignore[assignment]
         self.default_filename = default_filename
 
-    def compute_etag(self):
+    def compute_etag(self) -> str | None:
         """Compute the etag."""
         return None
 
     @classmethod
-    def get_absolute_path(cls, roots, path):
+    def get_absolute_path(cls, roots: Sequence[str], path: str) -> str:
         """locate a file to serve on our static file search path"""
         with cls._lock:
             if path in cls._static_paths:
@@ -980,9 +1019,9 @@ class FileFindHandler(JupyterHandler, web.StaticFileHandler):
             log().debug(f"Path {path} served from {abspath}")
             return abspath
 
-    def validate_absolute_path(self, root, absolute_path):
+    def validate_absolute_path(self, root: str, absolute_path: str) -> str | None:
         """check if the file should be served (raises 404, 403, etc.)"""
-        if absolute_path == "":
+        if not absolute_path:
             raise web.HTTPError(404)
 
         for root in self.root:
@@ -995,7 +1034,9 @@ class FileFindHandler(JupyterHandler, web.StaticFileHandler):
 class APIVersionHandler(APIHandler):
     """An API handler for the server version."""
 
-    def get(self):
+    _track_activity = False
+
+    def get(self) -> None:
         """Get the server version info."""
         # not authenticated, so give as few info as possible
         self.finish(json.dumps({"version": jupyter_server.__version__}))
@@ -1007,7 +1048,7 @@ class TrailingSlashHandler(web.RequestHandler):
     This should be the first, highest priority handler.
     """
 
-    def get(self):
+    def get(self) -> None:
         """Handle trailing slashes in a get."""
         assert self.request.uri is not None
         path, *rest = self.request.uri.partition("?")
@@ -1023,7 +1064,7 @@ class TrailingSlashHandler(web.RequestHandler):
 class MainHandler(JupyterHandler):
     """Simple handler for base_url."""
 
-    def get(self):
+    def get(self) -> None:
         """Get the main template."""
         html = self.render_template("main.html")
         self.write(html)
@@ -1035,7 +1076,7 @@ class FilesRedirectHandler(JupyterHandler):
     """Handler for redirecting relative URLs to the /files/ handler"""
 
     @staticmethod
-    async def redirect_to_files(self, path):
+    async def redirect_to_files(self: Any, path: str) -> None:
         """make redirect logic a reusable static method
 
         so it can be called from other handlers.
@@ -1063,19 +1104,19 @@ class FilesRedirectHandler(JupyterHandler):
         self.log.debug("Redirecting %s to %s", self.request.path, url)
         self.redirect(url)
 
-    def get(self, path=""):
-        return self.redirect_to_files(self, path)
+    async def get(self, path: str = "") -> None:
+        return await self.redirect_to_files(self, path)
 
 
 class RedirectWithParams(web.RequestHandler):
     """Sam as web.RedirectHandler, but preserves URL parameters"""
 
-    def initialize(self, url, permanent=True):
+    def initialize(self, url: str, permanent: bool = True) -> None:
         """Initialize a redirect handler."""
         self._url = url
         self._permanent = permanent
 
-    def get(self):
+    def get(self) -> None:
         """Get a redirect."""
         sep = "&" if "?" in self._url else "?"
         url = sep.join([self._url, self.request.query])
@@ -1087,7 +1128,7 @@ class PrometheusMetricsHandler(JupyterHandler):
     Return prometheus metrics for this server
     """
 
-    def get(self):
+    def get(self) -> None:
         """Get prometheus metrics."""
         if self.settings["authenticate_prometheus"] and not self.logged_in:
             raise web.HTTPError(403)
@@ -1097,7 +1138,7 @@ class PrometheusMetricsHandler(JupyterHandler):
 
 
 # -----------------------------------------------------------------------------
-# URL pattern fragments for re-use
+# URL pattern fragments for reuse
 # -----------------------------------------------------------------------------
 
 # path matches any number of `/foo[/bar...]` or just `/` or ''
